from typing import Optional

from shimoku.api.resources.action import Action
from shimoku.api.resources.universe import Universe

from shimoku.exceptions import UniverseApiKeyError, ActionError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging

logger = logging.getLogger(__name__)


async def get_universe_api_key(
    universe: Universe, universe_api_key: Optional[str]
) -> str:
    """Get the universe API key or create one if none exists
    :param universe: Universe
    :param universe_api_key: Universe API key
    :return: Universe API key
    """
    universe_api_keys = await universe.get_universe_api_keys()
    if universe_api_key is None:
        if len(universe_api_keys) == 0:
            return (await universe.create_universe_api_key("Actions key"))["id"]
        return universe_api_keys[0]["id"]
    if universe_api_key not in [k["id"] for k in universe_api_keys]:
        log_error(
            logger,
            f"Universe API key {universe_api_key} not found in {universe_api_keys}",
            UniverseApiKeyError,
        )
    return universe_api_key


class ActionsLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the Workspace layer
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(
        self,
        universe: Universe,
    ):
        self._universe = universe

    # @add_to_general_async_group
    async def create_action(
        self,
        name: str,
        description: Optional[str] = None,
        action_template_id: Optional[str] = None,
        universe_api_key: Optional[str] = None,
        path_to_code: Optional[str] = None,
        code: Optional[str] = None,
        overwrite: Optional[bool] = False,
        libraries: Optional[list[str]] = None,
    ) -> str:
        """
        Create an action.
        :param name: name of the action
        :param description: description of the action
        :param action_template_id: id of the action template
        :param universe_api_key: id of the universe API key
        :param path_to_code: path to the code
        :param code: code of the action
        :param libraries: libraries of the action
        :param overwrite: whether to overwrite the action if it already exists
        """
        if not libraries:
            libraries = []
        if code and path_to_code:
            log_error(
                logger,
                "You cannot provide both a path to code and code",
                ActionError,
            )
        if path_to_code:
            with open(path_to_code, "r") as file:
                code = file.read()
        if not code:
            log_error(
                logger,
                "Code is empty, make sure to provide a path to code or code itself",
                ActionError,
            )
        universe_api_key = await get_universe_api_key(
            universe=self._universe, universe_api_key=universe_api_key
        )
        if overwrite:
            action = await self._universe.get_action(name=name)
            if action:
                await self._universe.delete_action(uuid=action["id"])
        action: Action = await self._universe.create_action(
            name=name,
            description=description,
            action_template_id=action_template_id,
            universe_api_key_id=universe_api_key,
            libraries=libraries,
        )
        await action.upload_code(code)
        logger.info(
            f"Action ({name}) created with id ({action['id']}) and code uploaded"
        )
        return action["id"]

    # @add_to_individual_async_group
    async def update_action(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        new_description: Optional[str] = None,
    ):
        """
        Update an action.
        :param uuid: uuid of the action
        :param name: name of the action
        :param new_name: new name of the action
        :param new_description: new description of the action
        """
        await self._universe.update_action(
            uuid=uuid, name=name, new_name=new_name, new_description=new_description
        )

    # @add_to_individual_async_group
    async def delete_action(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete an action.
        :param uuid: uuid of the action
        :param name: name of the action
        """
        await self._universe.delete_action(uuid=uuid, name=name)
        logger.info(f"Action ({name or uuid}) deleted")

    async def get_action_metadata(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Get an action.
        :param uuid: uuid of the action
        :param name: name of the action
        :return: The action
        """
        action = await self._universe.get_action(uuid=uuid, name=name)
        if not action:
            log_error(
                logger,
                f"Action ({name}) not found",
                ActionError,
            )
        action_as_dict = action.cascade_to_dict()
        del action_as_dict["scripts"]
        return action_as_dict

    async def get_action_code(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        """
        Get the code of the action.
        :param uuid: uuid of the action
        :param name: name of the action
        :return: The code of the action
        """
        action = await self._universe.get_action(uuid=uuid, name=name)
        if not action:
            log_error(
                logger,
                f"Action ({name}) not found",
                ActionError,
            )
        return await action.get_code()



from shimoku import Client
from .data.stacked_area import stacked_area_data


def stacked_area_chart(shimoku_client: Client):
    print("test_area_chart")
    shimoku_client.set_menu_path("test-free-echarts", "stacked-area-chart")

    shimoku_client.plt.stacked_area(
        data=stacked_area_data,
        x="Weekday",
        x_axis_name="Visits per weekday",
        order=0,
    )



import dataclasses
import inspect
import importlib
import os
from shimoku.utils import get_args_with_defaults
import subprocess

base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
target_paths = {
    os.path.join(base_path, "src", "shimoku", "api", "user_access_classes"): "*",
    os.path.join(base_path, "src", "shimoku", "ai"): "ai_layer.py",
    os.path.join(base_path, "src", "shimoku", "plt"): "plt_layer.py",
}


def handle_annotation(annotation: type) -> str:
    if hasattr(annotation, "_use_info_logging") and annotation._use_info_logging:
        return f"{annotation.__name__}Header"
    if annotation is type(None) or annotation is None:
        type_str = "None"
    elif hasattr(annotation, "__name__"):
        type_str = annotation.__name__
    else:
        type_str = str(annotation)
    return (
        type_str.replace("NoneType", "None")
        .replace("pandas.core.frame.DataFrame", " DataFrame")
        .replace("typing.", "")
        .replace("<built-in function callable>", "callable")
    )


def handle_parameter(name: str, annotation: type, default_args: dict) -> str:
    if name in default_args:
        default_value = default_args[name]
        if isinstance(default_value, str):
            default_value = f'"{default_value}"'
        return f"        {name}: {handle_annotation(annotation)} = {default_value},"
    else:
        return f"        {name}: {handle_annotation(annotation)},"


def handle_method(method) -> list[str]:
    method_code_lines = []
    method_code_lines.extend(
        [
            f"    def {method[0]}(",
            "        self,",
            *[
                handle_parameter(
                    arg,
                    method[1].__annotations__[arg],
                    get_args_with_defaults(method[1]),
                )
                for arg in method[1].__annotations__
                if arg not in ["self", "return"]
            ],
        ]
    )
    sig = inspect.signature(method[1])
    params = sig.parameters.values()
    has_kwargs = any([True for p in params if p.kind == p.VAR_KEYWORD])
    if has_kwargs:
        method_code_lines.append("        **kwargs,")
    if "return" in method[1].__annotations__:
        method_code_lines.append(
            f'    ) -> {handle_annotation(method[1].__annotations__["return"])}:'
        )
    else:
        method_code_lines.append("    ):")
    method_code_lines.extend(
        [
            f'        """{method[1].__doc__}"""',
            "        pass",
            "",
        ]
    )
    return method_code_lines


def handle_class(class_for_generation) -> list[str]:
    print(class_for_generation[0])
    is_dataclass = hasattr(class_for_generation[1], "__dataclass_fields__")

    fields_code_lines = []
    if is_dataclass:
        for field in class_for_generation[1].__dataclass_fields__.values():
            if field.default is dataclasses.MISSING:
                fields_code_lines.append(f"    {field.name}: {handle_annotation(field.type)}")
            else:
                fields_code_lines.append(
                    f"    {field.name}: {handle_annotation(field.type)} = {field.default}"
                )

    methods = inspect.getmembers(
        class_for_generation[1], predicate=inspect.isfunction
    )
    methods_code_lines = []
    for method in methods:
        if method[0].startswith("_"):
            continue
        methods_code_lines.extend(handle_method(method))

    inner_classes = inspect.getmembers(
        class_for_generation[1], predicate=inspect.isclass
    )

    inner_classes_code_lines = []
    for inner_class in inner_classes:
        if inner_class[0].startswith("_"):
            continue
        inner_classes_code_lines.extend(["    "+line for line in handle_class(inner_class)])

    if is_dataclass:
        class_header_code_lines = [
            f"@dataclass",
            f"class {class_for_generation[0]}:",
            '    """',
            *["    " + line for line in fields_code_lines],
            '    """',
        ]
    else:
        class_header_code_lines = [
            f"class {class_for_generation[0]}Header:",
            f'    """{class_for_generation[1].__doc__}"""',
            ""
        ]

    class_code_lines = [
        "",
        *class_header_code_lines,
        *fields_code_lines,
        "",
        *inner_classes_code_lines,
        *methods_code_lines,
    ]
    return class_code_lines


def main():
    for target_path, target_file in target_paths.items():
        output_path = os.path.join(target_path, "generated_headers")
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        for file in os.listdir(target_path):
            if file != target_file and target_file != "*":
                continue
            if file.endswith(".py") and not file.startswith("__"):
                print(
                    f'{os.path.relpath(target_path, base_path).replace(os.sep, ".")[4:]}.{file[:-3]}'
                )
                module = importlib.import_module(
                    f'{os.path.relpath(target_path, base_path).replace(os.sep, ".")[4:]}.{file[:-3]}'
                )
                classes_for_generation = []
                for name, obj in inspect.getmembers(module):
                    if not inspect.isclass(obj):
                        continue
                    if name.startswith("_"):
                        continue
                    if not hasattr(obj, "_use_info_logging") or not obj._use_info_logging:
                        continue
                    classes_for_generation.append((name, obj))

                if not classes_for_generation:
                    continue

                classes_code_lines = []
                for class_for_generation in classes_for_generation:
                    classes_code_lines.extend(handle_class(class_for_generation))

                main_code_lines = [
                    "# This file was generated automatically by "
                    "scripts/user_classes_header_generator.py do not modify it!",
                    "# If the user access files are modified, this file has to be regenerated with the script.",
                    "from typing import Optional, Union",
                    "from pandas import DataFrame",
                    "from dataclasses import dataclass",
                    "import shimoku",
                    "",
                    *classes_code_lines,
                ]
                with open(
                    os.path.join(output_path, f"{classes_for_generation[-1][0]}Header.py"),
                    "w",
                ) as f:
                    f.write("\n".join(main_code_lines))

                subprocess.run(
                    [
                        "black",
                        os.path.join(
                            output_path, f"{classes_for_generation[-1][0]}Header.py"
                        ),
                    ]
                )


if __name__ == "__main__":
    main()
